import numpy as np



class Rbf:
    def __init__(self, coordinates, function_values, kernel, epsilon = 1, norm_weights = None):
        self.coordinates = np.asarray(coordinates)
        self.function_values = np.asarray(function_values)

        self.n = len(coordinates)
        self.kernel = kernel
        self.epsilon = epsilon
        self.weights = np.zeros(len(coordinates))

        if norm_weights != None and len(norm_weights) == len(coordinates[0]):
            self.__norm_weights = norm_weights
        elif norm_weights != None and len(norm_weights) != len(coordinates[0]):
            print("Improper norm_weights shape, defaulted to Euclidean norm.")
            self.__norm_weights = np.full((len(self.coordinates[0])), 1)
        else:
            self.__norm_weights = np.full((len(self.coordinates[0])), 1)

        self.coefficient_matrix = np.empty((self.n, self.n))
        self.is_singular = False
        Rbf.__BuildInterpolant(self)
        self.average_error = 0
        self.errors = Rbf.LeaveOneOutError(self)

    def Norm(self, v):
        norm = 0
        for i in range(len(v)):
            norm += (self.__norm_weights[i]*v[i])**2
        return np.sqrt(norm)
    
    def MinBoundingNBallFromOrigin(self):
        max = np.zeros_like(self.coordinates[0])
        
        for coord in self.coordinates:
            for i, x in enumerate(coord):
                if max[i] < np.abs(x):
                    max[i] = np.abs(x)
        return 2*np.linalg.norm(max, 2)

    def __BuildInterpolant(self):
        for i in range(self.n):
            for j in range(i, self.n, 1):
                r = Rbf.Norm(self, self.coordinates[i] - self.coordinates[j])
                self.coefficient_matrix[i][j] = Rbf.__kernel(self, r)
                if i != j:
                    self.coefficient_matrix[j][i] = Rbf.__kernel(self, r)
        try:
            self.weights = np.linalg.solve(self.coefficient_matrix, self.function_values)
        except:
            print("ill-conditioned matrix (coefficient matrix not full rank)")
            self.is_singular = True
            return

    def Interpolate(self, point):
        out = 0
        for i in range(self.n):
            r = Rbf.Norm(self, self.coordinates[i] - point)
            out += Rbf.__kernel(self, r)*self.weights[i]
        return out

    #error is rms error
    def LeaveOneOutError(self):
        if self.is_singular:
            print("cannot compute error, interpolation matrix is singular")
            self.average_error = np.inf
            return
        inverse = np.linalg.inv(self.coefficient_matrix)
        error = [abs(self.weights[i]/inverse[i][i])**2 for i in range(self.n)]
        self.average_error = np.sqrt(abs(np.average(error)))
        return np.asarray(error)
    
    def ShapeOpitmizer(self, epsilon):
        self.epsilon = epsilon
        Rbf.__BuildInterpolant(self)
        error = Rbf.LeaveOneOutError(self)
        return self.average_error

    def __kernel(self, r):
        if self.kernel == "linear":
            return self.epsilon*r
        elif self.kernel == "cubic":
            return self.epsilon*r**3
        elif self.kernel == "thin_plate_spline":
            if r == 0:
                return 0
            else:
                return self.epsilon*r**2 * np.log(r)
        elif self.kernel == "quintic":
            return self.epsilon*r**5
        elif self.kernel == "gaussian":
            return np.exp(-1.0*(self.epsilon*r)**2)
        elif self.kernel == "multiquadric":
            return -1.0*np.sqrt(1 + (self.epsilon*r)**2)
        elif self.kernel == "inverse_quadric":
            return 1/(1 + (self.epsilon*r)**2)
    


from scipy.interpolate import RBFInterpolator
import numpy as np
from distutils import errors
from lazyfit.Rbf import Rbf
from scipy.optimize import minimize_scalar

class LazyFitter:
    def __init__(self, coordinates, function_values, epsilon = 1):
        self.coordinates = np.asarray(coordinates)
        self.function_values = np.asarray(function_values)

        self.min_error = []
        self.epsilon = epsilon
        self.errors = []
        LazyFitter.Fit(self)
        self.interpolator = Rbf(self.coordinates, self.function_values, kernel = self.min_error[0], epsilon=self.epsilon)
    
    def Fit(self):
        kernels = ["linear", "cubic", "thin_plate_spline", "quintic", "gaussian", "multiquadric", "inverse_quadric"]
        for kernel in kernels:
            LazyFitter.TestKernel(self, kernel)
        min = self.errors[0]
        for error_pack in self.errors:
            if error_pack[1] < min[1]:
                min = error_pack
        self.min_error = min

    def OptimizeShape(self):
        self.interpolator = Rbf(self.coordinates, self.function_values, kernel = self.min_error[0])
        best_epsilon = minimize_scalar(self.interpolator.ShapeOpitmizer, bounds = (0.000001, self.interpolator.MinBoundingNBallFromOrigin()))
        print(best_epsilon.success)
        self.epsilon = best_epsilon.x
        self.interpolator = Rbf(self.coordinates, self.function_values, kernel = self.min_error[0], epsilon=best_epsilon.x)
        self.min_error[1] = self.interpolator.average_error

    def TestKernel(self, kernel):
        interp = Rbf(self.coordinates, self.function_values, kernel)
        self.errors.append([kernel, interp.average_error])
    
    def Interpolate(self, x):
        return self.interpolator.Interpolate(x)


from setuptools import setup

######################################################################################################
################ You May Remove All the Comments Once You Finish Modifying the Script ################
######################################################################################################

setup(
    
    name = 'lazyfit', 
    
    version = '0.1.2',
    
    description = 'A python package that helps with the construction and optimization of RBF interpolants.',
    
    package_dir = {'':'src'},
    
    packages = ['lazyfit'],
    
    author = 'Alex Shaffer',
    author_email = 'alex.shaffer144@gmail.com',
    
    
    long_description = open('README.md').read() + '\n\n' + open('CHANGELOG.md').read(),
    long_description_content_type = "text/markdown",
    
    url='https://github.com/redhorn144/lazyfit.git',
    
    
    include_package_data=True,
    
    classifiers  = [
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: OS Independent',
    ],
    
    install_requires = [

        'numpy ~= 1.23.5',
        'scipy ~= 1.10.1'

    ],
    
    keywords = ['Interpolation', 'Data Science'],
    
    )
